"""
SSSD LDAP provider tests

:requirement: IDM-SSSD-REQ : LDAP Provider
"""

from __future__ import annotations

import time

import pytest
from sssd_test_framework.roles.client import Client
from sssd_test_framework.roles.ldap import LDAP
from sssd_test_framework.topology import KnownTopology


@pytest.mark.ticket(bz=[795044, 1695574])
@pytest.mark.importance("critical")
@pytest.mark.parametrize("modify_mode", ["exop", "ldap_modify", "exop_force"])
@pytest.mark.parametrize("use_ppolicy", ["true", "false"])
@pytest.mark.parametrize("sssd_service_user", ("root", "sssd"))
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.require(
    lambda client, sssd_service_user: ((sssd_service_user == "root") or client.features["non-privileged"]),
    "SSSD was built without support for running under non-root",
)
@pytest.mark.builtwith("ldap_use_ppolicy")
def test_ldap__password_change_using_ppolicy(
    client: Client, ldap: LDAP, modify_mode: str, use_ppolicy: str, sssd_service_user: str
):
    """
    :title: Password change using ppolicy
    :description: PPolicy overlay is the latest implementation of IETF password policy for LDAP.
    This extends the password policy for the LDAP server and is configured in SSSD using
    'ldap_use_ppolicy'.

    Two password modification modes are tested, Extended Operation (exop), the default and then
    LDAP (ldapmodify), set by 'ldap_pwmodify_mode' parameter.
    :note: This feature is introduced in SSSD 2.10.0
    :setup:
        1. Add a user to LDAP
        2. Configure the LDAP ACI to permit user password changes
        3. Set "ldap_pwmodify_mode"
        4. Start SSSD
    :steps:
        1. Authenticate as user
        2. Change the password of user
        3. Authenticate user with new password
        4. Authenticate user with old password
    :expectedresults:
        1. User is authenticated
        2. Password is changed successfully
        3. User is authenticated
        4. User is not authenticated
    :customerscenario: True
    """
    user = "user1"
    old_pass = "Secret123"
    new_pass = "New_password123"

    ldap.user(user).add(password=old_pass)
    ldap.aci.add('(targetattr="userpassword")(version 3.0; acl "pwp test"; allow (all) userdn="ldap:///self";)')

    client.sssd.domain["ldap_pwmodify_mode"] = modify_mode
    client.sssd.domain["ldap_use_ppolicy"] = use_ppolicy
    client.sssd.start(service_user=sssd_service_user)

    assert client.auth.ssh.password(user, old_pass), "Login with old password failed!"

    assert client.auth.passwd.password(user, old_pass, new_pass), "Password change failed!"

    assert client.auth.ssh.password(user, new_pass), "User login failed!"
    assert not client.auth.ssh.password(user, old_pass), "Login with old password passed!"


@pytest.mark.ticket(bz=[795044, 1695574])
@pytest.mark.importance("critical")
@pytest.mark.parametrize("modify_mode", ["exop", "ldap_modify", "exop_force"])
@pytest.mark.parametrize("use_ppolicy", ["true", "false"])
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.builtwith("ldap_use_ppolicy")
def test_ldap__password_change_new_passwords_do_not_match_using_ppolicy(
    client: Client, ldap: LDAP, modify_mode: str, use_ppolicy: str
):
    """
    :title: Password change when the new passwords do not match
    :setup:
        1. Add user to LDAP
        2. Configure the LDAP ACI to permit user password changes
        3. set "ldap_pwmodify_mode"
        4. Start SSSD
    :steps:
        1. Change password to new password, but retyped password is different
    :expectedresults:
        1. Password change is not successful
    :customerscenario: True
    """
    ldap.user("user1").add(password="Secret123")
    ldap.aci.add('(targetattr="userpassword")(version 3.0; acl "pwp test"; allow (all) userdn="ldap:///self";)')

    client.sssd.domain["ldap_pwmodify_mode"] = modify_mode
    client.sssd.domain["ldap_use_ppolicy"] = use_ppolicy
    client.sssd.start()

    assert not client.auth.passwd.password(
        "user1", "Secret123", "Red123", "Hat000"
    ), "Password should not have been able to be changed!"


@pytest.mark.ticket(bz=[795044, 1695574, 1795220])
@pytest.mark.importance("critical")
@pytest.mark.parametrize("modify_mode", ["exop", "ldap_modify", "exop_force"])
@pytest.mark.parametrize("use_ppolicy", ["true", "false"])
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.builtwith("ldap_use_ppolicy")
def test_ldap__password_change_new_password_does_not_meet_complexity_requirements_using_ppolicy(
    client: Client, ldap: LDAP, modify_mode: str, use_ppolicy: str
):
    """
    :title: Password change when the new passwords do not meet the complexity requirements using ppolicy
    :setup:
        1. Add a user to LDAP
        2. Configure the LDAP ACI to permit user password changes
        3. Set "passwordCheckSyntax" to "on"
        4. Set "ldap_pwmodify_mode"
        5. Start SSSD
    :steps:
        1. Change password to new password, but all letters are lower-case
        2. Check logs
    :expectedresults:
        1. Password change failed
        2. Password change failure is logged
    :customerscenario: True
    """
    ldap.user("user1").add(password="Secret123")
    ldap.aci.add('(targetattr="userpassword")(version 3.0; acl "pwp test"; allow (all) userdn="ldap:///self";)')
    ldap.ldap.modify("cn=config", replace={"passwordCheckSyntax": "on"})

    client.sssd.domain["ldap_pwmodify_mode"] = modify_mode
    client.sssd.domain["ldap_use_ppolicy"] = use_ppolicy
    client.sssd.start()

    assert not client.auth.passwd.password(
        "user1", "Secret123", "red_32"
    ), "Password should not have been able to be changed!"

    match = client.journald.is_match(r"pam_sss\(passwd:chauthtok\): User info message: Password change failed.")
    assert match, "'Password change failed.' message is not in log!"


@pytest.mark.ticket(bz=[1695574, 1795220])
@pytest.mark.importance("critical")
@pytest.mark.parametrize("modify_mode", ["exop", "ldap_modify", "exop_force"])
@pytest.mark.parametrize("use_ppolicy", ["true", "false"])
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.builtwith("ldap_use_ppolicy")
def test_ldap__password_change_with_invalid_current_password_using_ppolicy(
    client: Client, ldap: LDAP, modify_mode: str, use_ppolicy: str
):
    """
    :title: Password change fails with invalid current password
    :setup:
        1. Add a user to LDAP, set his password
        2. Configure the LDAP ACI to permit user password changes
        3. Set "ldap_pwmodify_mode"
        4. Start SSSD
    :steps:
        1. Attempt to change the password but enter the incorrect password
    :expectedresults:
        1. Password change is not successful
    :customerscenario: True
    """
    ldap.user("user1").add(password="Secret123")
    ldap.aci.add('(targetattr="userpassword")(version 3.0; acl "pwp test"; allow (all) userdn="ldap:///self";)')

    client.sssd.domain["ldap_pwmodify_mode"] = modify_mode
    client.sssd.domain["ldap_use_ppolicy"] = use_ppolicy
    client.sssd.start()

    assert not client.auth.passwd.password(
        "user1", "wrong123", "Newpass123"
    ), "Password should not have been able to be changed!"


@pytest.mark.importance("low")
@pytest.mark.ticket(bz=[1067476, 1065534])
@pytest.mark.topology(KnownTopology.LDAP)
def test_ldap__authenticate_user_with_whitespace_prefix_in_userid(client: Client, ldap: LDAP):
    """
    :title: Authenticate with a user containing a blank space in the userid
    :description: This can only be tested on LDAP because most directories have
    constraints on the values, protecting the integrity of the data. This scenario
    is most likely created migrating and, or upgrading old databases.
    :setup:
        1. Add users " space1" and "user1"
        2. Start SSSD
    :steps:
        1. Lookup user " space1"
        2. Login user " space1"
        3. Login user "space1"
        4. Lookup "user1" user
        5. Lookup " user1" user
    :expectedresults:
        1. " space1" is found and has correct id
        2. " space1" is able to log in
        3. "space1" is not able to log in
        4. "user1" is found and has correct id
        5. " user1" is not found
    :customerscenario: True
    """
    ldap.user(" space1").add(uid=10011, password="Secret123")
    ldap.user("user1").add(uid=10012, password="Secret123")
    client.sssd.start()

    result = client.tools.id(" space1")
    assert result is not None, "User ' space1' was not found"
    assert result.user.id == 10011, "User ' space1' has wrong id"

    assert client.auth.ssh.password(" space1", "Secret123"), "User ' space1' login failed!"
    assert not client.auth.ssh.password("space1", "Secret123"), "User 'space1' login should have failed!"

    result = client.tools.id("user1")
    assert result is not None, "User 'user1' not found!"
    assert result.user.id == 10012, "User 'user1' has the wrong uid!"

    result = client.tools.id(" user1")
    assert result is None, "User ' user1' should not be found!"


@pytest.mark.importance("high")
@pytest.mark.ticket(bz=1507035)
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.parametrize("method", ["su", "ssh"])
def test_ldap__change_password_when_ldap_pwd_policy_is_set_to_shadow(client: Client, ldap: LDAP, method: str):
    """
    :title: Change password with shadow ldap password policy is set to shadow
    :description: Changing a password when the password policy is managed by the shadowAccount objectclass.
    :setup:
        1. Configure the LDAP ACI to permit user password changes
        2. Create user with shadowLastChange = 0, shadowMin = 0, shadowMax = 99999 and shadowWarning = 7
        3. Set "ldap_pwd_policy = shadow"
        4. Set "ldap_chpass_update_last_change = True"
        5. Start SSSD
    :steps:
        1. Authenticate as "tuser" with old password
        2. Authenticate as "tuser" with new password
    :expectedresults:
        1. The password is expired, and the user is forced to change their password
        2. Authentication with new password was successful
    :customerscenario: True
    """
    ldap.aci.add('(targetattr="userpassword")(version 3.0; acl "pwp test"; allow (all) userdn="ldap:///self";)')
    ldap.user("tuser").add(
        uid=999011, gid=999011, shadowMin=0, shadowMax=99999, shadowWarning=7, shadowLastChange=0, password="Secret123"
    )

    client.sssd.domain["ldap_pwd_policy"] = "shadow"
    client.sssd.domain["ldap_chpass_update_last_change"] = "True"
    client.sssd.start()

    assert client.auth.parametrize(method).password_expired(
        "tuser", "Secret123", "Redhat@321"
    ), "Password change failed!"
    assert client.auth.parametrize(method).password("tuser", "Redhat@321"), "User 'tuser' login failed!"

    log = client.fs.read(f"/var/log/sssd/sssd_{client.sssd.default_domain}.log")
    for timeout in ["ldap_opt_timeout", "ldap_search_timeout", "ldap_network_timeout", "dns_resolver_timeout"]:
        assert timeout in log, f"Value '{timeout}' not found in logs"


@pytest.mark.importance("critical")
@pytest.mark.topology(KnownTopology.LDAP)
def test_ldap__search_base_is_discovered_and_defaults_to_root_dse(client: Client, ldap: LDAP):
    """
    :title: Search base is discovered and defaults to the directories root DSE
    :setup:
        1. Create OU
        2. Create user, and put the user object in the new OU
        3. Start SSSD
    :steps:
        1. Authenticate as user and check the logs
        2. Add "ldap_search_base" to the configuration and cleanly restart SSSD
        3. Authenticate as user and check the logs
    :expectedresults:
        1. User authentication is successful, and logs contain messages discovering root DSE
        2. SSSD is configured and cleanly restarted
        3. User authentication is successful, and logs contain no messages searching root DSE
    :customerscenario: False
    """
    base = ldap.ldap.naming_context

    ou_users = ldap.ou("users").add()
    user = ldap.user("puser1", basedn=ou_users).add(uid=10001, gid=10001, password="Secret123")

    client.sssd.start()

    assert client.auth.ssh.password(user.name, "Secret123")
    time.sleep(3)

    log = client.fs.read(client.sssd.logs.domain())
    for doc in [
        f"Setting option [ldap_search_base] to [{base}]",
        f"Setting option [ldap_user_search_base] to [{base}]",
        f"Setting option [ldap_group_search_base] to [{base}]",
        f"Setting option [ldap_netgroup_search_base] to [{base}]",
    ]:
        assert doc in str(log), f"String '{doc}' not found in logs!"
    client.sssd.dom("test")["ldap_search_base"] = ldap.ldap.naming_context

    client.sssd.stop()
    client.sssd.clear()
    client.sssd.start()

    assert client.auth.ssh.password("puser1", "Secret123"), "User 'puser1' login failed!"
    time.sleep(3)

    log = client.fs.read(client.sssd.logs.domain())
    assert "sdap_set_config_options_with_rootdse" not in log, "sdap_set_config_options_with_rootdse found in logs!"


@pytest.mark.importance("medium")
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.parametrize(
    "user_search_base, search_base",
    [
        ("ldap_user_search_base", "ou=People,dc=ldap,dc=test"),
        ("ldap_group_search_base", "ou=Groups,dc=ldap,dc=test"),
        ("ldap_netgroup_search_base", "ou=Netgroup,dc=ldap,dc=test"),
    ],
)
def test_ldap__search_base_is_discovered_and_defaults_to_root_dse_users_groups_and_netgroups(
    client: Client, ldap: LDAP, user_search_base, search_base
):
    """
    :title: Search base is discovered and defaults to the directories root DSE for users, groups and netgroups
    :setup:
        1. Create People OU
        2. Create user, and put the user object in the new OU
        3. Configure SSSD with "user_search_base" and start SSSD
    :steps:
        1. Lookup user and authenticate as user and check the logs
    :expectedresults:
        1. User authentication is successful, and logs contain messages setting ldap_x_search_base to the root DSE
    :customerscenario: False
    """
    base = ldap.ldap.naming_context
    ou_users = ldap.ou("People").add()
    user = ldap.user("puser1", basedn=ou_users).add(uid=10001, gid=10001, password="Secret123")

    client.sssd.dom("test")[user_search_base] = search_base
    client.sssd.start()

    result = client.tools.getent.passwd(user.name)
    assert result is not None
    assert result.name == user.name

    assert client.auth.ssh.password(user.name, "Secret123")
    time.sleep(3)

    log = client.fs.read(client.sssd.logs.domain())
    match user_search_base:
        case "ldap_user_search_base":
            for doc in [
                "Got rootdse",
                f"Setting option [ldap_search_base] to [{base}]",
                f"Setting option [ldap_group_search_base] to [{base}]",
                f"Setting option [ldap_netgroup_search_base] to [{base}]",
            ]:
                assert doc in str(log), f"String '{doc}' not found in logs!"
        case "ldap_group_search_base":
            for doc in [
                "Got rootdse",
                f"Setting option [ldap_search_base] to [{base}]",
                f"Setting option [ldap_user_search_base] to [{base}]",
                f"Setting option [ldap_netgroup_search_base] to [{base}]",
            ]:
                assert doc in str(log), f"String '{doc}' not found in logs!"
        case "ldap_netgroup_search_base":
            for doc in [
                "Got rootdse",
                f"Setting option [ldap_search_base] to [{base}]",
                f"Setting option [ldap_user_search_base] to [{base}]",
                f"Setting option [ldap_group_search_base] to [{base}]",
            ]:
                assert doc in str(log), f"String '{doc}, not found in logs!"


@pytest.mark.topology(KnownTopology.LDAP)
def test_ldap__lookup_user_with_search_bases(client: Client, ldap: LDAP):
    """
    :title: Looking up with no search base configured
    :setup:
        1. Create People OU
        2. Create user and put the user object into the OU
        3. Start SSSD
    :steps:
        1. Lookup user
        2. Look at logs
    :expectedresults:
        1. User is found
        2. Strings pertaining to rootdse search base discovery are found
    :customerscenario: False
    """
    base = ldap.ldap.naming_context

    ou_users = ldap.ou("People").add()
    user = ldap.user("puser1", basedn=ou_users).add(uid=10001, gid=10001, password="Secret123")

    client.sssd.start()

    result = client.tools.getent.passwd(user.name)
    assert result is not None, "User not found!"
    assert result.name == user.name, "Username is not correct!"
    time.sleep(3)

    log = client.fs.read(client.sssd.logs.domain())
    assert "Got rootdse" in log, "Unable to find rootDSE!"
    assert "Using value from [defaultNamingContext] as naming context" in log, "Unable to find naming context!"
    assert f"Setting option [ldap_search_base] to [{base}]" in log, "Unable to set ldap_search_base!"


@pytest.mark.importance("low")
@pytest.mark.topology(KnownTopology.LDAP)
@pytest.mark.parametrize("user_search_base", ["dc=ldap,dc=test", "dc=shanks,dc=com"])
def test_ldap__lookup_and_authenticate_as_user_with_different_object_search_bases(
    client: Client, ldap: LDAP, user_search_base
):
    """
    :title: Looking up and authenticating as a user when the default, user and group search bases are different
    :setup:
        1. Create People OU
        2. Create user and put the user object into the new OU
        3. Configure "ldap_search_base", "ldap_user|group_search_base" and start SSSD
    :steps:
        1. Lookup and authenticate as user
    :expectedresults:
        1. User lookup and authentication are successful
    :customerscenario: False
    """
    base = ldap.ldap.naming_context

    ou_users = ldap.ou("People").add()
    user = ldap.user("puser1", basedn=ou_users).add(uid=10001, gid=10001, password="Secret123")

    client.sssd.dom("test")["ldap_search_base"] = user_search_base
    client.sssd.dom("test")["ldap_user_search_base"] = f"ou=People,{base}"
    client.sssd.dom("test")["ldap_group_search_base"] = f"ou=Groups,{base}"

    client.sssd.start()

    result = client.tools.getent.passwd(user.name)
    assert result is not None, "User is not found!"
    assert result.name == user.name, "Username is not correct!"
    assert client.auth.ssh.password(user.name, "Secret123"), "User login failed!"


@pytest.mark.ticket(jira="RHEL-55993")
@pytest.mark.importance("critical")
@pytest.mark.parametrize(
    "modify_mode, expected, err_msg",
    [("exop", 1, "Expected login failure"), ("exop_force", 3, "Expected password change request")],
)
@pytest.mark.parametrize("method", ["su", "ssh"])
@pytest.mark.topology(KnownTopology.LDAP)
def test_ldap__password_change_no_grace_logins_left(
    client: Client, ldap: LDAP, modify_mode: str, expected: int, err_msg: str, method: str
):
    """
    :title: Password change when no grace logins left
    :description: Typically the LDAP extended operation to change a password
    requires an authenticated bind, even if the data send with the extended
    operation contains the old password. If the old password is expired and
    there are no grace logins left an authenticated bind is not possible anymore
    and as a result it is not possible for the user to change their password.
    With 'exop' SSSD will not try to ask the user for new credentials while with
    'exop_force' SSSD will ask for new credentials and will try to run the password
    change extended operation.
    :setup:
        1. Set "passwordExp" to "on"
        2. Set "passwordMaxAge" to "1"
        3. Set "passwordGraceLimit" to "0"
        4. Add a user to LDAP
        5. Wait until the password is expired
        6. Set "ldap_pwmodify_mode"
        7. Start SSSD
    :steps:
        1. Authenticate as the user with 'exop_force' set
        2. Authenticate as the user with 'exop' set
    :expectedresults:
        1. With 'exop_force' expect a request to change the password
        2. With 'exop' expect just a failed login
    :customerscenario: False
    """
    ldap.ldap.modify("cn=config", replace={"passwordExp": "on", "passwordMaxAge": "1", "passwordGraceLimit": "0"})
    ldap.user("user1").add(password="Secret123")

    # make sure the password is expired
    time.sleep(3)

    client.sssd.domain["ldap_pwmodify_mode"] = modify_mode
    client.sssd.start()

    rc, _, _, _ = client.auth.parametrize(method).password_with_output("user1", "Secret123")
    assert rc == expected, err_msg



# Configuration file for multihost tests.

from __future__ import annotations

from pytest_mh import MultihostPlugin
from sssd_test_framework.config import SSSDMultihostConfig

# Load additional plugins
pytest_plugins = (
    "pytest_importance",
    "pytest_mh",
    "pytest_ticket",
    "sssd_test_framework.fixtures",
    "sssd_test_framework.markers",
)


def pytest_plugin_registered(plugin) -> None:
    if isinstance(plugin, MultihostPlugin):
        plugin.config_class = SSSDMultihostConfig



"""Bug 1268902 SSSD doesn't set the ID mapping range automatically

:requirement: IDM-SSSD-REQ: Active Directory host keytab renewal
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
import re
import time
import pytest
from sssd.testlib.common.utils import sssdTools


@pytest.mark.usefixtures('joinad')
@pytest.mark.idmaprange
class Testidmaprange(object):
    """ Test cases for BZ: 1268902

    :setup:
      1. Join to AD using adcli command.
      2. Add the user using adcli user add.
    :bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1268902
    """
    @pytest.mark.tier1
    def test_001_findrid(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Find RID
         value from objectSID
        :id: ecb488b2-7694-4e68-9ee7-8833cb7100b1
        """
        (_, rid) = get_rid
        assert rid != 0 or rid is not None

    @pytest.mark.tier1
    def test_002_rangelessthansid(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Verify user
         lookup when ldap idmap range size less than sids
        :id: de6144c2-ac5a-4b00-a7e1-bd6a87f99134
        """
        (ad_user, rid) = get_rid
        new_rid = str(rid - 1)
        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'ldap_idmap_range_size': new_rid,
                       'debug_level': '9'}
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        lookup_cmd = 'getent passwd %s' % ad_user
        cmd = multihost.client[0].run_command(lookup_cmd, raiseonerr=False)
        time.sleep(5)
        if cmd.returncode == 0:
            rid = client.find_rid(ad_user)
            assert rid != 0 or rid is not None
        client.sssd_conf(dom_section, sssd_params, action='delete')

    @pytest.mark.tier2
    def test_003_disablerange(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Verify user
         lookup when ldap idmap range size less than sids with disable feature
        :id: 13d9a69a-de96-4884-872d-579191740cc0
        """
        (ad_user, rid) = get_rid
        assert rid != 0 or rid is not None
        client = sssdTools(multihost.client[0], multihost.ad[0])
        domain_name = client.get_domain_section_name()
        new_rid = str(rid - 1)
        multihost.client[0].service_sssd('stop')
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'ldap_idmap_range_size': new_rid,
                       'ldap_idmap_helper_table_size': '0',
                       'debug_level': '9'}
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        log = ("/var/log/sssd/sssd_%s.log" % domain_name)
        lookup_cmd = 'getent passwd %s' % ad_user
        getent = multihost.client[0].run_command(lookup_cmd, raiseonerr=False)
        time.sleep(15)
        if not getent.returncode == 0:
            log_1 = re.compile(r'.*RID\sthat\sis\slarger\sthan\sthe\s'
                               r'ldap_idmap_range_size.\sSee\sthe\s"ID\s'
                               r'MAPPING"\ssection\sof\ssssd-ad.5.\sfor\san.*')
            test_str_log = multihost.client[0].get_file_contents(log)
            search = log_1.search(test_str_log.decode())
            if search:
                print(search.group(0))
                assert True
            else:
                assert False
        client.sssd_conf(dom_section, sssd_params, action='delete')

    @pytest.mark.tier2
    def test_004_rangeequalsid(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Verify user
         lookup when ldap idmap range size equal to sids
        :id: 3687ca8c-5b40-434e-b699-0598b1db19e4
        """
        (ad_user, new_rid) = get_rid
        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'ldap_idmap_range_size': str(new_rid),
                       'debug_level': '9'}
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        lookup_cmd = 'getent passwd %s' % ad_user
        log = ("/var/log/sssd/sssd_%s.log" % domain_name)
        cmd = multihost.client[0].run_command(lookup_cmd, raiseonerr=False)
        user = ad_user.split('@')[0]
        time.sleep(15)
        if cmd.returncode == 0:
            log_1 = re.compile(r'Mapping\suser\s.%s.*' % user)
            test_str_log = multihost.client[0].get_file_contents(log)
            result = log_1.search(test_str_log.decode())
            if result:
                print(result.group(0))
                assert True
            else:
                assert False
        client.sssd_conf(dom_section, sssd_params, action='delete')

    @pytest.mark.tier2
    def test_005_disablerange(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Verify user
         lookup when ldap idmap range size equal to sids with disable feature
        :id: 6682d1c8-2ea8-47a2-9d67-3d48e0949816
        """
        (ad_user, new_rid) = get_rid
        multihost.client[0].service_sssd('stop')
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'ldap_idmap_range_size': str(new_rid),
                       'ldap_idmap_helper_table_size': '0',
                       'debug_level': '9'}
        client.sssd_conf(dom_section, sssd_params)
        client.clear_sssd_cache()
        time.sleep(15)
        log = ("/var/log/sssd/sssd_%s.log" % domain_name)
        lookup_cmd = 'getent passwd %s' % ad_user
        cmd = multihost.client[0].run_command(lookup_cmd, raiseonerr=False)
        time.sleep(5)
        if not cmd.returncode == 0:
            log_1 = re.compile(r'.*RID\sthat\sis\slarger\sthan\sthe\s'
                               r'ldap_idmap_range_size.\sSee\sthe\s"ID\s'
                               r'MAPPING"\ssection\sof\ssssd-ad.5.\sfor\san.*')
            test_str_log = multihost.client[0].get_file_contents(log)
            result = log_1.search(test_str_log.decode())
            if result:
                print(result.group(0))
                assert True
            else:
                assert False
        client.sssd_conf(dom_section, sssd_params, action='delete')

    @pytest.mark.tier2
    def test_006_rangevalues(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Verify user
         lookup when ldap idmap range size near border values
        :id: 54178ed0-10b4-4d23-b134-c8c99ebbd724
        """
        rid_list = []
        (ad_user, rid) = get_rid
        rid_list.append(str(rid - 2))
        rid_list.append(str(rid + 1))
        rid_list.append(str(rid + 2))
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        dom_section = 'domain/%s' % domain_name
        sssd_params = {'debug_level': '9'}
        for rid_val in rid_list:
            multihost.client[0].service_sssd('stop')
            sssd_params['ldap_idmap_range_size'] = rid_val
            sssd_params['debug_level'] = '9'
            client.sssd_conf(dom_section, sssd_params)
            client.clear_sssd_cache()
            time.sleep(15)
            log = ("/var/log/sssd/sssd_%s.log" % domain_name)
            lookup_cmd = 'getent passwd %s' % ad_user
            cmd = multihost.client[0].run_command(lookup_cmd, raiseonerr=False)
            user = ad_user.split('@')[0]
            if cmd.returncode == 0:
                log_1 = re.compile(r'Mapping\suser\s.%s.*' % user)
                test_str_log = multihost.client[0].get_file_contents(log)
                result = log_1.search(test_str_log.decode())
                if result:
                    print(result.group(0))
                    assert True
                else:
                    assert False
        client.sssd_conf(dom_section, sssd_params, action='delete')

    @pytest.mark.tier2
    def test_007_disablerangevalues(self, multihost, get_rid):
        """
        :title: IDM-SSSD-TC: Support large AD RIDs automatically: Verify user
         lookup when ldap idmap range size near border values with disable
         feature
        :id: 58e85bda-9b03-45c6-b508-5d3bb56ea6cf
        """
        rid_list = []
        (ad_user, rid) = get_rid
        rid_list.append(str(rid - 2))
        rid_list.append(str(rid + 1))
        rid_list.append(str(rid + 2))
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        section = 'domain/%s' % domain_name
        sssd_params = {'ldap_idmap_helper_table_size': '0',
                       'debug_level': '9'}
        for rid_val in rid_list:
            multihost.client[0].service_sssd('stop')
            sssd_params['ldap_idmap_range_size'] = rid_val
            client.sssd_conf(section, sssd_params)
            client.clear_sssd_cache()
            log = ("/var/log/sssd/sssd_%s.log" % domain_name)
            lookup_cmd = 'getent passwd %s' % ad_user
            user = ad_user.split('@')[0]
            cmd = multihost.client[0].run_command(lookup_cmd, raiseonerr=False)
            time.sleep(5)
            if not cmd.returncode == 0:
                log_1 = re.compile(r'.*RID\sthat\sis\slarger\sthan\sthe\s'
                                   r'ldap_idmap_range_size.\sSee\sthe\s"ID\s'
                                   r'MAPPING"\ssection\sof\ssssd-ad.5.\s'
                                   r'for\san.*')
            else:
                log_1 = re.compile(r'Mapping\suser\s.%s.*' % user)
            test_str_log = multihost.client[0].get_file_contents(log)
            result = log_1.search(test_str_log.decode())
            if result:
                print(result.group(0))
                assert True
            else:
                assert False
        client.sssd_conf(section, sssd_params, action='delete')



'''
Created on Sep 18, 2009

@author: sgallagh
'''

import os
import re
import shutil
import subprocess
import sys
from contextlib import suppress
from .sssdoptions import SSSDOptions
from .ipachangeconf import SSSDChangeConf

# Exceptions
class SSSDConfigException(Exception): pass
class ParsingError(Exception): pass
class AlreadyInitializedError(SSSDConfigException): pass
class NotInitializedError(SSSDConfigException): pass
class NoOutputFileError(SSSDConfigException): pass
class NoServiceError(SSSDConfigException): pass
class NoSectionError(SSSDConfigException): pass
class NoOptionError(SSSDConfigException): pass
class ServiceNotRecognizedError(SSSDConfigException): pass
class ServiceAlreadyExists(SSSDConfigException): pass
class NoDomainError(SSSDConfigException): pass
class DomainNotRecognized(SSSDConfigException): pass
class DomainAlreadyExistsError(SSSDConfigException): pass
class NoSuchProviderError(SSSDConfigException): pass
class NoSuchProviderSubtypeError(SSSDConfigException): pass
class ProviderSubtypeInUse(SSSDConfigException): pass

def striplist(l):
    return([x.strip() for x in l])

def options_overlap(options1, options2):
    overlap = []
    for option in options1:
        if option in options2:
            overlap.append(option)
    return overlap

class SSSDConfigSchema(SSSDChangeConf):
    def __init__(self, schemafile, schemaplugindir):
        SSSDChangeConf.__init__(self)
        #TODO: get these from a global setting
        if not schemafile:
            schemafile = '${prefix}/share/sssd/sssd.api.conf'
        if not schemaplugindir:
            schemaplugindir = '${prefix}/share/sssd/sssd.api.d'

        try:
            # Read the primary config file
            with open(schemafile, 'r') as fd:
                self.readfp(fd)
            # Read in the provider files
            for file in filter(lambda f: re.search(r'^sssd-.*\.conf$', f),
                                         os.listdir(schemaplugindir)):
                with open(schemaplugindir+ "/" + file) as fd:
                    self.readfp(fd)
        except IOError:
            raise
        except SyntaxError: # can be raised with readfp
            raise ParsingError

        # Set up lookup table for types
        self.type_lookup = {
            'bool' : bool,
            'int'  : int,
            'long' : long if sys.version_info[0] == 2 else int,
            'float': float,
            'str'  : str,
            'list' : list,
            'None' : None
            }

        # Lookup table for acceptable boolean values
        self.bool_lookup = {
            'false' : False,
            'true'  : True,
            }

    def get_options(self, section):
        if not self.has_section(section):
            raise NoSectionError
        options = self.options(section)

        # Indexes
        PRIMARY_TYPE = 0
        SUBTYPE = 1
        MANDATORY = 2
        DEFAULT = 3

        # Parse values
        parsed_options = {}
        for option in self.strip_comments_empty(options):
            unparsed_option = option['value']
            split_option = striplist(unparsed_option.split(','))
            optionlen = len(split_option)

            primarytype = self.type_lookup[split_option[PRIMARY_TYPE]]
            subtype = self.type_lookup[split_option[SUBTYPE]]
            mandatory = self.bool_lookup[split_option[MANDATORY]]

            if option['name'] in SSSDOptions.option_strings:
                desc = SSSDOptions.option_strings[option['name']]
            else:
                desc = None

            if optionlen == 3:
                # This option has no defaults
                parsed_options[option['name']] = \
                    (primarytype,
                     subtype,
                     mandatory,
                     desc,
                     None)
            elif optionlen == 4:
                if type(split_option[DEFAULT]) == primarytype:
                    parsed_options[option['name']] = \
                        (primarytype,
                         subtype,
                         mandatory,
                         desc,
                         split_option[DEFAULT])
                elif primarytype == list:
                    if (type(split_option[DEFAULT]) == subtype):
                        parsed_options[option['name']] = \
                            (primarytype,
                             subtype,
                             mandatory,
                             desc,
                             [split_option[DEFAULT]])
                    else:
                        try:
                            if subtype == bool and \
                            type(split_option[DEFAULT]) == str:
                                parsed_options[option['name']] = \
                                    (primarytype,
                                     subtype,
                                     mandatory,
                                     desc,
                                     [self.bool_lookup[split_option[DEFAULT].lower()]])
                            else:
                                parsed_options[option['name']] = \
                                    (primarytype,
                                     subtype,
                                     mandatory,
                                     desc,
                                     [subtype(split_option[DEFAULT])])
                        except ValueError:
                            raise ParsingError
                else:
                    try:
                        if primarytype == bool and \
                            type(split_option[DEFAULT]) == str:
                                parsed_options[option['name']] = \
                                    (primarytype,
                                     subtype,
                                     mandatory,
                                     desc,
                                     self.bool_lookup[split_option[DEFAULT].lower()])
                        else:
                            parsed_options[option['name']] = \
                                (primarytype,
                                 subtype,
                                 mandatory,
                                 desc,
                                 primarytype(split_option[DEFAULT]))
                    except ValueError:
                        raise ParsingError

            elif optionlen > 4:
                if (primarytype != list):
                    raise ParsingError
                fixed_options = []
                for x in split_option[DEFAULT:]:
                    if type(x) != subtype:
                        try:
                            if (subtype == bool and type(x) == str):
                                newvalue = self.bool_lookup[x.lower()]
                            else:
                                newvalue = subtype(x)
                            fixed_options.extend([newvalue])
                        except ValueError:
                            raise ParsingError
                    else:
                        fixed_options.extend([x])
                parsed_options[option['name']] = \
                    (primarytype,
                     subtype,
                     mandatory,
                     desc,
                     fixed_options)
            else:
                # Bad config file
                raise ParsingError

        return parsed_options

    def get_option(self, section, option):
        if not self.has_section(section):
            raise NoSectionError(section)
        if not self.has_option(section, option):
            raise NoOptionError("Section [%s] has no option [%s]" %
                                (section, option))

        return self.get_options(section)[option]

    def get_defaults(self, section):
        if not self.has_section(section):
            raise NoSectionError(section)

        schema_options = self.get_options(section)
        defaults = dict([(x,schema_options[x][4])
                         for x in schema_options.keys()
                         if schema_options[x][4] != None])

        return defaults

    def get_services(self):
        service_list = [x['name'] for x in self.sections()
                        if x['name'] != 'service' and
                        not x['name'].startswith('domain') and
                        not x['name'].startswith('provider')]
        return service_list

    def get_providers(self):
        providers = {}
        for section in self.sections():
            splitsection = section['name'].split('/')
            if (splitsection[0] == 'provider'):
                if(len(splitsection) == 3):
                    if splitsection[1] not in providers:
                        providers[splitsection[1]] = []
                    providers[splitsection[1]].extend([splitsection[2]])
        for key in providers.keys():
            providers[key] = tuple(providers[key])
        return providers

class SSSDConfigObject(object):
    def __init__(self):
        self.name = None
        self.options = {}

    def get_name(self):
        """
        Return the name of the object

        === Returns ===
        The domain name

        === Errors ===
        No errors
        """
        return self.name

    def get_option(self, optionname):
        """
        Return the value of an service option

        optionname:
          The option to get.

        === Returns ===
        The value for the requested option.

        === Errors ===
        NoOptionError:
          The specified option was not listed in the service
        """
        if optionname in self.options.keys():
            return self.options[optionname]
        raise NoOptionError(optionname)

    def get_all_options(self):
        """
        Return a dictionary of name/value pairs for this object

        === Returns ===
        A dictionary of name/value pairs currently in use for this object

        === Errors ===
        No errors
        """
        return self.options

    def remove_option(self, optionname):
        """
        Remove an option from the object. If the option does not exist, it is ignored.

        === Returns ===
        No return value.

        === Errors ===
        No errors
        """
        if optionname in self.options:
            del self.options[optionname]

class SSSDService(SSSDConfigObject):
    '''
    Object to manipulate SSSD service options
    '''

    def __init__(self, servicename, apischema):
        """
        Create a new SSSDService, setting its defaults to those found in the
        schema. This constructor should not be used directly. Use
        SSSDConfig.new_service() instead.

        name:
          The service name
        apischema:
          An SSSDConfigSchema? object created by SSSDConfig.__init__()

        === Returns ===
        The newly-created SSSDService object.

        === Errors ===
        TypeError:
          The API schema passed in was unusable or the name was not a string.
        ServiceNotRecognizedError:
          The service was not listed in the schema
        """
        SSSDConfigObject.__init__(self)

        if not isinstance(apischema, SSSDConfigSchema) or type(servicename) != str:
            raise TypeError

        if not apischema.has_section(servicename):
            raise ServiceNotRecognizedError(servicename)

        self.name = servicename
        self.schema = apischema

        # Set up the service object with any known defaults
        self.options = {}

        # Include a list of hidden options
        self.hidden_options = []

        # Set up default options for all services
        self.options.update(self.schema.get_defaults('service'))

        # Set up default options for this service
        self.options.update(self.schema.get_defaults(self.name))

    def list_options_with_mandatory(self):
        """
        List options for the service, including the mandatory flag.

        === Returns ===
        A dictionary of configurable options. This dictionary is keyed on the
        option name with a tuple of the variable type, subtype ('None' if the
        type is not  a collection type), whether it is mandatory, the
        translated option description, and the default value (or 'None') as
        the value.

        Example:
        { 'enumerate' :
          (bool, None, False, u'Enable enumerating all users/groups', True) }

        === Errors ===
        No errors
        """
        options = {}

        # Get the list of available options for all services
        schema_options = self.schema.get_options('service')
        options.update(schema_options)

        schema_options = self.schema.get_options(self.name)
        options.update(schema_options)

        return options

    def list_options(self):
        """
        List all options that apply to this service

        === Returns ===
        A dictionary of configurable options. This dictionary is keyed on the
        option name with a tuple of the variable type, subtype ('None' if the
        type is not  a collection type), the translated option description, and
        the default value (or 'None') as the value.

        Example:
        { 'services' :
          (list, str, u'SSSD Services to start', ['nss', 'pam']) }

        === Errors ===
        No Errors
        """
        options = self.list_options_with_mandatory()

        # Filter out the mandatory field to maintain compatibility
        # with older versions of the API
        filtered_options = {}
        for key in options.keys():
            filtered_options[key] = (options[key][0], options[key][1], options[key][3], options[key][4])

        return filtered_options

    def list_mandatory_options(self):
        """
        List all mandatory options that apply to this service

        === Returns ===
        A dictionary of configurable options. This dictionary is keyed on the
        option name with a tuple of the variable type, subtype ('None' if the
        type is not  a collection type), the translated option description, and
        the default value (or 'None') as the value.

        Example:
        { 'services' :
          (list, str, u'SSSD Services to start', ['nss', 'pam']) }

        === Errors ===
        No Errors
        """
        options = self.list_options_with_mandatory()

        # Filter out the mandatory field to maintain compatibility
        # with older versions of the API
        filtered_options = {}
        for key in options.keys():
            if options[key][2]:
                filtered_options[key] = (options[key][0], options[key][1], options[key][3], options[key][4])

        return filtered_options

    def set_option(self, optionname, value):
        """
        Set a service option to the specified value (or values)

        optionname:
          The option to change
        value:
          The value to set. This may be a single value or a list of values. If
          it is set to None, it resets the option to its default.

        === Returns ===
        No return value

        === Errors ===
        NoOptionError:
          The specified option is not listed in the schema
        TypeError:
          The value specified was not of the expected type
        """
        if self.schema.has_option(self.name, optionname):
            option_schema = self.schema.get_option(self.name, optionname)
        elif self.schema.has_option('service', optionname):
            option_schema = self.schema.get_option('service', optionname)
        elif optionname in self.hidden_options:
            # Set this option and do not add it to the list of changeable values
            self.options[optionname] = value
            return
        else:
            raise NoOptionError('Section [%s] has no option [%s]' % (self.name, optionname))

        if value == None:
            self.remove_option(optionname)
            return

        raise_error = False

        # If we were expecting a list and didn't get one,
        # Create a list with a single entry. If it's the
        # wrong subtype, it will fail below
        if option_schema[0] == list and type(value) != list:
            if type(value) == str:
                value = striplist(value.split(','))
            else:
                value = [value]

        if type(value) != option_schema[0]:
            # If it's possible to convert it, do so
            try:
                if option_schema[0] == bool and type(value) == str:
                    value = self.schema.bool_lookup[value.lower()]
                elif option_schema[0] == int and type(value) == str:
                    # Make sure we handle any reasonable base
                    value = int(value, 0)
                else:
                    value = option_schema[0](value)
            except ValueError:
                raise_error = True
            except KeyError:
                raise_error = True

            if raise_error:
                raise TypeError('Expected %s for %s, received %s' %
                                (option_schema[0], optionname, type(value)))

        if type(value) == list:
            # Iterate through the list an ensure that all members
            # are of the appropriate subtype
            try:
                newvalue = []
                for x in value:
                    if option_schema[1] == bool and \
                    type(x) == str:
                        newvalue.extend([self.schema.bool_lookup[x.lower()]])
                    else:
                        newvalue.extend([option_schema[1](x)])
            except ValueError:
                raise_error = True
            except KeyError:
                raise_error = True

            if raise_error:
                raise TypeError('Expected %s' % option_schema[1])

            value = newvalue

        self.options[optionname] = value

class SSSDDomain(SSSDConfigObject):
    """
    Object to manipulate SSSD domain options
    """
    def __init__(self, domainname, apischema):
        """
        Creates a new, empty SSSDDomain. This domain is inactive by default.
        This constructor should not be used directly. Use
        SSSDConfig.new_domain() instead.

        name:
          The domain name.
        apischema:
          An SSSDConfigSchema object created by SSSDConfig.__init__()

        === Returns ===
        The newly-created SSSDDomain object.

        === Errors ===
        TypeError:
          apischema was not an SSSDConfigSchema object or domainname was not
         a string
        """
        SSSDConfigObject.__init__(self)

        if not isinstance(apischema, SSSDConfigSchema) or type(domainname) != str:
            raise TypeError

        self.name = domainname
        self.schema = apischema
        self.active = False
        self.oldname = None
        self.providers = []

        # Set up the domain object with any known defaults
        self.options = {}

        # Set up default options for all domains
        self.options.update(self.schema.get_defaults('provider'))
        self.options.update(self.schema.get_defaults('domain'))

    def set_active(self, active):
        """
        Enable or disable this domain

        active:
          Boolean value. If True, this domain will be added to the active
          domains list when it is saved. If False, it will be removed from the
          active domains list when it is saved.

        === Returns ===
        No return value

        === Errors ===
        No errors
        """
        self.active = bool(active)

    def list_options_with_mandatory(self):
        """
        List options for the currently-configured providers, including the
        mandatory flag

        === Returns ===
        A dictionary of configurable options. This dictionary is keyed on the
        option name with a tuple of the variable type, subtype ('None' if the
        type is not  a collection type), whether it is mandatory, the
        translated option description, and the default value (or 'None') as
        the value.

        Example:
        { 'enumerate' :
          (bool, None, False, u'Enable enumerating all users/groups', True) }

        === Errors ===
        No errors
        """
        options = {}
        # Get the list of available options for all domains
        options.update(self.schema.get_options('provider'))

        options.update(self.schema.get_options('domain'))

        # Candidate for future optimization: will update primary type
        # for each subtype
        for (provider, providertype) in self.providers:
            schema_options = self.schema.get_options('provider/%s'
                                                     % provider)
            options.update(schema_options)
            schema_options = self.schema.get_options('provider/%s/%s'
                                                     % (provider, providertype))
            options.update(schema_options)
        return options

    def list_options(self):
        """
        List options available for the currently-configured providers.

        === Returns ===
        A dictionary of configurable options. This dictionary is keyed on the
        option name with a tuple of the variable type, subtype ('None' if the
        type is not  a collection type), the translated option description, and
        the default value (or 'None') as the value.

        Example:
        { 'enumerate' :
          (bool, None, u'Enable enumerating all users/groups', True) }

        === Errors ===
        No errors
        """
        options = self.list_options_with_mandatory()

        # Filter out the mandatory field to maintain compatibility
        # with older versions of the API
        filtered_options = {}
        for key in options.keys():
            filtered_options[key] = (options[key][0], options[key][1], options[key][3], options[key][4])

        return filtered_options

    def list_mandatory_options(self):
        """
        List mandatory options for the currently-configured providers.

        === Returns ===
        A dictionary of configurable options. This dictionary is keyed on the
        option name with a tuple of the variable type, subtype ('None' if the
        type is not  a collection type), the translated option description, and
        the default value (or 'None') as the value.

        Example:
        { 'enumerate' :
          (bool, None, u'Enable enumerating all users/groups', True) }

        === Errors ===
        No errors
        """
        options = self.list_options_with_mandatory()

        # Filter out the mandatory field to maintain compatibility
        # with older versions of the API
        filtered_options = {}
        for key in options.keys():
            if options[key][2]:
                filtered_options[key] = (options[key][0], options[key][1], options[key][3], options[key][4])

        return filtered_options

    def list_provider_options(self, provider, provider_type=None):
        """
        If provider_type is specified, list all options applicable to that
        target, otherwise list all possible options available for a provider.

        type:
            Provider backend type. (e.g. local, ldap, krb5, etc.)
        provider_type:
            Subtype of the backend type. (e.g. id, auth, access, chpass)

        === Returns ===

        A dictionary of configurable options for the specified provider type.
        This dictionary is keyed on the option name with a tuple of the
        variable type, subtype ('None' if the type is not  a collection type),
        the translated option description, and the default value (or 'None')
        as the value.

        === Errors ===

        NoSuchProviderError:
            The specified provider is not listed in the schema or plugins
        NoSuchProviderSubtypeError:
            The specified provider subtype is not listed in the schema
        """
        #TODO section checking

        options = self.schema.get_options('provider/%s' % provider)
        if(provider_type):
            options.update(self.schema.get_options('provider/%s/%s' %
                                                   (provider, provider_type)))
        else:
            # Add options from all provider subtypes
            known_providers = self.list_providers()
            for provider_type in known_providers[provider]:
                options.update(self.list_provider_options(provider,
                                                          provider_type))
        return options

    def list_providers(self):
        """
        Return a dictionary of providers.

        === Returns ===
        Returns a dictionary of providers, keyed on the primary type, with the
        value being a tuple of the subtypes it supports.

        Example:
        { 'ldap' : ('id', 'auth', 'chpass') }

        === Errors ===
        No Errors
        """
        return self.schema.get_providers()

    def set_option(self, option, value):
        """
        Set a domain option to the specified value (or values)

        option:
          The option to change.
        value:
          The value to set. This may be a single value or a list of values.
          If it is set to None, it resets the option to its default.

        === Returns ===
        No return value.

        === Errors ===
        NoOptionError:
            The specified option is not listed in the schema
        TypeError:
            The value specified was not of the expected type
        """
        options = self.list_options()
        if (option not in options.keys()):
            raise NoOptionError('Section [%s] has no option [%s]' %
                                (self.name, option))

        if value == None:
            self.remove_option(option)
            return

        option_schema = options[option]
        raise_error = False

        # If we were expecting a list and didn't get one,
        # Create a list with a single entry. If it's the
        # wrong subtype, it will fail below
        if option_schema[0] == list and type(value) != list:
            if type(value) == str:
                value = striplist(value.split(','))
            else:
                value = [value]

        if type(value) != option_schema[0]:
            # If it's possible to convert it, do so
            try:
                if option_schema[0] == bool and \
                type(value) == str:
                    value = self.schema.bool_lookup[value.lower()]
                elif option_schema[0] == int and type(value) == str:
                    # Make sure we handle any reasonable base
                    value = int(value, 0)
                else:
                    value = option_schema[0](value)
            except ValueError:
                raise_error = True
            except KeyError:
                raise_error = True

            if raise_error:
                raise TypeError('Expected %s for %s, received %s' %
                                (option_schema[0], option, type(value)))

        if type(value) == list:
            # Iterate through the list an ensure that all members
            # are of the appropriate subtype
            try:
                newvalue = []
                for x in value:
                    if option_schema[1] == bool and \
                    type(x) == str:
                        newvalue.extend([self.schema.bool_lookup[x.lower()]])
                    else:
                        newvalue.extend([option_schema[1](x)])
            except ValueError:
                raise_error = True
            except KeyError:
                raise_error = True

            if raise_error:
                raise TypeError('Expected %s' % option_schema[1])
            value = newvalue

        # Check whether we're adding a provider entry.
        is_provider = option.rfind('_provider')
        if (is_provider > 0):
            provider = option[:is_provider]
            try:
                self.add_provider(value, provider)
            except NoSuchProviderError:
                raise NoOptionError
        else:
            self.options[option] = value

    def set_name(self, newname):
        """
        Change the name of the domain

        newname:
          New name for this domain

        === Returns ===
        No return value.

        === Errors ===
        TypeError:
          newname was not a string
        """

        if type(newname) != str:
            raise TypeError

        if not self.oldname:
            # Only set the oldname once
            self.oldname = self.name
        self.name = newname

    def add_provider(self, provider, provider_type):
        """
        Add a new provider type to the domain

        type:
          Provider backend type. (e.g. local, ldap, krb5, etc.)
        subtype:
          Subtype of the backend type. (e.g. id, auth, chpass)

        === Returns ===
        No return value.

        === Errors ===
        ProviderSubtypeInUse:
          Another backend is already providing this subtype
        NoSuchProviderError:
          The specified provider is not listed in the schema or plugins
        NoSuchProviderSubtypeError:
          The specified provider subtype is not listed in the schema
        """
        # Check that provider and provider_type are valid
        configured_providers = self.list_providers()
        if provider in configured_providers.keys():
            if provider_type not in configured_providers[provider]:
                raise NoSuchProviderSubtypeError(provider_type)
        else:
            raise NoSuchProviderError

        # Don't add a provider twice
        with_this_type = [x for x in self.providers if x[1] == provider_type]
        if len(with_this_type) > 1:
            # This should never happen!
            raise ProviderSubtypeInUse
        if len(with_this_type) == 1:
            if with_this_type[0][0] != provider:
                raise ProviderSubtypeInUse(with_this_type[0][0])
        else:
            self.providers.extend([(provider, provider_type)])

        option_name = '%s_provider' % provider_type
        self.options[option_name] = provider

        # Add defaults for this provider
        self.options.update(self.schema.get_defaults('provider/%s' %
                                                     provider))
        self.options.update(self.schema.get_defaults('provider/%s/%s' %
                                                     (provider,
                                                      provider_type)))

    def remove_provider(self, provider_type):
        """
        Remove a provider from the domain. If the provider is not present, it
        is ignored.

        provider_type:
          Subtype of the backend type. (e.g. id, auth, chpass)

        === Returns ===
        No return value.

        === Errors ===
        No Errors
        """

        provider = None
        for (provider, ptype) in self.providers:
            if ptype == provider_type:
                break
            provider = None

        # Check whether the provider_type was found
        if not provider:
            return

        # Remove any unused options when removing the provider.
        options = self.list_provider_options(provider, provider_type)

        # Trim any options that are used by other providers,
        # if that provider is in use
        for (prov, ptype) in self.providers:
            # Ignore the one being removed
            if (prov, ptype) == (provider, provider_type):
                continue

            provider_options = self.list_provider_options(prov, ptype)
            overlap = options_overlap(options.keys(), provider_options.keys())
            for opt in overlap:
                del options[opt]

        # We should now have a list of options used only by this
        # provider. So we remove them.
        for option in options:
            if option in self.options:
                del self.options[option]

        # Remove this provider from the option list
        option = '%s_provider' % provider_type
        if option in self.options:
            del self.options[option]

        self.providers.remove((provider, provider_type))

class SSSDConfig(SSSDChangeConf):
    """
    class SSSDConfig
    Primary class for operating on SSSD configurations
    """
    def __init__(self, schemafile=None, schemaplugindir=None):
        """
        Initialize the SSSD config parser/editor. This constructor does not
        open or create a config file. If the schemafile and schemaplugindir
        are not passed, it will use the system defaults.

        schemafile:
          The path to the API schema config file. Usually
          ${prefix}/share/sssd/sssd.api.conf
        schemaplugindir:
          The path the directory containing the provider schema config files.
          Usually ${prefix}/share/sssd/sssd.api.d

        === Returns ===
        The newly-created SSSDConfig object.

        === Errors ===
        IOError:
          Exception raised when the schema file could not be opened for
          reading.
        ParsingError:
          The main schema file or one of those in the plugin directory could
          not be parsed.
        """
        SSSDChangeConf.__init__(self)
        self.schema = SSSDConfigSchema(schemafile, schemaplugindir)
        self.configfile = None
        self.initialized = False
        self.API_VERSION = 2

    def import_config(self,configfile=None):
        """
        Read in a config file, populating all of the service and domain
        objects with the read values.

        configfile:
          The path to the SSSD config file. If not specified, use the system
          default, usually ${prefix}/etc/sssd.conf

        === Returns ===
        No return value

        === Errors ===
        IOError:
          Exception raised when the file could not be opened for reading
        ParsingError:
          Exception raised when errors occur attempting to parse a file.
        AlreadyInitializedError:
          This SSSDConfig object was already initialized by a call to
          import_config() or new_config()
        """
        if self.initialized:
            raise AlreadyInitializedError

        if not configfile:
            #TODO: get this from a global setting
            configfile = '${prefix}/etc/sssd/sssd.conf'
        # open will raise an IOError if it fails
        with open(configfile, 'r') as fd:
            try:
                self.readfp(fd)
            except Exception:
                raise ParsingError

        self.configfile = configfile
        self.initialized = True

    def new_config(self):
        """
        Initialize the SSSDConfig object with the defaults from the schema.

        === Returns ===
        No return value

        === Errors ===
        AlreadyInitializedError:
          This SSSDConfig object was already initialized by a call to
          import_config() or new_config()
        """
        if self.initialized:
            raise AlreadyInitializedError

        self.initialized = True

        #Initialize all services
        for servicename in self.schema.get_services():
            service = self.new_service(servicename)

    def write(self, outputfile=None):
        """
        Write out the configuration to a file.

        outputfile:
          The path to write the new config file. If it is not specified, it
          will use the path specified by the import() call.
        === Returns ===
        No return value

        === Errors ===
        IOError:
          Exception raised when the file could not be opened for writing
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        NoOutputFileError:
          No outputfile was specified and this SSSDConfig object was not
          initialized by import()
        """
        if not self.initialized:
            raise NotInitializedError

        if outputfile == None:
            if(self.configfile == None):
                raise NoOutputFileError

            outputfile = self.configfile

        # open() will raise IOError if it fails
        old_umask = os.umask(0o137)
        with open(outputfile, "w") as of:
            output = self.dump(self.opts)
            of.write(output)
        os.umask(old_umask)
        service_user = ""
        ret = subprocess.run(["systemctl", "show", "sssd", "--value", "--property", "User"], capture_output=True, text=True)
        if ret.returncode == 0:
            service_user = ret.stdout.strip()
        if service_user == "":
            service_user = "root"
        with suppress(PermissionError):
            shutil.chown(outputfile, "root", service_user)

    def list_active_services(self):
        """
        Return a list of all active services.

        === Returns ===
        The list of active services.

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError

        if (self.has_option('sssd', 'services')):
            active_services = striplist(self.get('sssd', 'services').split(','))
            service_dict = dict.fromkeys(active_services)
            if '' in service_dict:
                del service_dict['']

            # Remove any entries in this list that don't
            # correspond to an active service, for integrity
            configured_services = self.list_services()
            for srv in list(service_dict):
                if srv not in configured_services:
                    del service_dict[srv]

            active_services = list(service_dict)
        else:
            active_services = []

        return active_services

    def list_inactive_services(self):
        """
        Return a list of all disabled services.

        === Returns ===
        The list of inactive services.

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError

        if (self.has_option('sssd', 'services')):
            active_services = striplist(self.get('sssd', 'services').split(','))
        else:
            active_services = []

        services = [x for x in self.list_services()
                   if x not in active_services]
        return services

    def list_services(self):
        """
        Retrieve a list of known services.

        === Returns ===
        The list of known services.

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError

        service_list = [x['name'] for x in self.sections()
                        if not x['name'].startswith('domain') ]
        return service_list

    def get_service(self, name):
        """
        Get an SSSDService object to edit a service.

        name:
          The name of the service to return.

        === Returns ===
        An SSSDService instance containing the current state of a service in
        the SSSDConfig

        === Errors ===
        NoServiceError:
          There is no such service with the specified name in the SSSDConfig.
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError
        if not self.has_section(name):
            raise NoServiceError

        service = SSSDService(name, self.schema)
        for opt in self.strip_comments_empty(self.options(name)):
            try:
                service.set_option(opt['name'], opt['value'])
            except NoOptionError:
                # If we come across an option that we don't recognize,
                # we should just ignore it and continue
                pass

        return service

    def new_service(self, name):
        """
        Create a new service from the defaults and return the SSSDService
        object for it. This function will also add this service to the list of
        active services in the [SSSD] section.

        name:
          The name of the service to create and return.

        === Returns ===
        The newly-created SSSDService object

        === Errors ===
        ServiceNotRecognizedError:
          There is no such service in the schema.
        ServiceAlreadyExistsError:
          The service being created already exists in the SSSDConfig object.
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError
        if (self.has_section(name)):
            raise ServiceAlreadyExists(name)

        service = SSSDService(name, self.schema)
        self.save_service(service)
        return service

    def activate_service(self, name):
        """
        Activate a service

        name:
          The name of the service to activate

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        NoServiceError:
          There is no such service with the specified name in the SSSDConfig.
        """

        if not self.initialized:
            raise NotInitializedError

        if name not in self.list_services():
            raise NoServiceError

        item = self.get_option_index('sssd', 'services')[1]
        if not item:
            self.set('sssd','services', name)
            return

        # Turn the items into a set of dictionary keys
        # This guarantees uniqueness and makes it easy
        # to add a new value
        service_dict = dict.fromkeys(striplist(item['value'].split(',')))
        if '' in service_dict:
            del service_dict['']

        # Add a new key for the service being activated
        service_dict[name] = None

        # Write out the joined keys
        self.set('sssd','services', ", ".join(service_dict.keys()))

    def deactivate_service(self, name):
        """
        Deactivate a service

        name:
          The name of the service to deactivate

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        NoServiceError:
          There is no such service with the specified name in the SSSDConfig.
        """

        if not self.initialized:
            raise NotInitializedError

        if name not in self.list_services():
            raise NoServiceError
        item = self.get_option_index('sssd', 'services')[1]
        if not item:
            self.set('sssd','services', '')
            return

        # Turn the items into a set of dictionary keys
        # This guarantees uniqueness and makes it easy
        # to remove the one unwanted value.
        service_dict = dict.fromkeys(striplist(item['value'].split(',')))
        if '' in service_dict:
            del service_dict['']

        # Remove the unwanted service from the lest
        if name in service_dict:
            del service_dict[name]

        # Write out the joined keys
        self.set('sssd','services', ", ".join(service_dict.keys()))

    def delete_service(self, name):
        """
        Remove a service from the SSSDConfig object. This function will also
        remove this service from the list of active services in the [SSSD]
        section. Has no effect if the service does not exist.

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError
        self.delete_option('section', name)

    def save_service(self, service):
        """
        Save the changes made to the service object back to the SSSDConfig
        object.

        service_object:
          The SSSDService object to save to the configuration.

        === Returns ===
        No return value
        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        TypeError:
          service_object was not of the type SSSDService
        """
        if not self.initialized:
            raise NotInitializedError
        if not isinstance(service, SSSDService):
            raise TypeError

        name = service.get_name()
        # Ensure that the existing section is removed
        # This way we ensure that we are getting a
        # complete copy of the service.
        # delete_option() is a noop if the section
        # does not exist.
        index = self.delete_option('section', name)

        addkw = []
        for option,value in service.get_all_options().items():
            if (type(value) == list):
                value = ', '.join(value)
            if option == "debug_level":
                value = self._get_debug_level_val(value)
            addkw.append( { 'type'  : 'option',
                            'name'  : option,
                            'value' : str(value) } )

        self.add_section(name, addkw, index)

    def list_active_domains(self):
        """
        Return a list of all active domains.

        === Returns ===
        The list of configured, active domains.

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError

        if (self.has_option('sssd', 'domains')):
            sssd_domains = striplist(self.get('sssd', 'domains').split(','))
            domain_dict = dict.fromkeys(sssd_domains)
            if '' in domain_dict:
                del domain_dict['']
            sssd_domains = list(domain_dict)
        else:
            sssd_domains = []

        domains = self.list_domains()
        for dom in self.list_domains():
            # Remove explicitly enabled from the list
            if self.has_option('domain/%s' % dom, 'enabled'):
                if self.get('domain/%s' % dom, 'enabled') == 'false':
                    domains.remove(dom)
                    if dom in sssd_domains:
                        sssd_domains.remove(dom)
            else:
                if dom not in sssd_domains:
                    domains.remove(dom)

        return domains

    def list_inactive_domains(self):
        """
        Return a list of all configured, but disabled domains.

        === Returns ===
        The list of configured, inactive domains.

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError

        if (self.has_option('sssd', 'domains')):
            sssd_domains = striplist(self.get('sssd', 'domains').split(','))
            domain_dict = dict.fromkeys(sssd_domains)
            if '' in domain_dict:
                del domain_dict['']
            sssd_domains = list(domain_dict)
        else:
            sssd_domains = []

        domains = self.list_domains()
        for dom in self.list_domains():
            # Remove explicitly enabled from the list
            if self.has_option('domain/%s' % dom, 'enabled'):
                if self.get('domain/%s' % dom, 'enabled') == 'true':
                    domains.remove(dom)
            else:
                if dom in sssd_domains:
                    domains.remove(dom)

        return domains

    def list_domains(self):
        """
        Return a list of all configured domains, including inactive domains.

        === Returns ===
        The list of configured domains, both active and inactive.

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError
        domains = [x['name'][7:] for x in self.sections() if x['name'].startswith('domain/')]
        return domains

    def get_domain(self, name):
        """
        Get an SSSDDomain object to edit a domain.

        name:
          The name of the domain to return.

        === Returns ===
        An SSSDDomain instance containing the current state of a domain in the
        SSSDConfig

        === Errors ===
        NoDomainError:
          There is no such domain with the specified name in the SSSDConfig.
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError
        if not self.has_section('domain/%s' % name):
            raise NoDomainError(name)

        domain = SSSDDomain(name, self.schema)

        # Read in the providers first or we may have type
        # errors trying to read in their options
        providers = [ (x['name'],x['value']) for x in self.strip_comments_empty(self.options('domain/%s' % name))
                     if x['name'].rfind('_provider') > 0]

        # Handle the default value for providers if those weren't set in the configuration file
        default_providers = {
            'id_provider': None,
            'sudo_provider': 'id_provider',
            'auth_provider': 'id_provider',
            'chpass_provider': 'auth_provider',
            'autofs_provider': 'id_provider',
            'selinux_provider': 'id_provider',
            'subdomains_provider': 'id_provider',
            'session_provider': 'id_provider',
            'hostid_provider': 'id_provider',
            'resolver_provider': 'id_provider',
        }

        configured_providers = domain.list_providers()
        providers_list = [x[0] for x in providers]

        if 'access_provider' not in providers_list:
            providers.append(('access_provider', 'permit'))

        for provider, default_provider in default_providers.items():
            if provider in providers_list:
                continue

            if default_provider in providers_list:
                default_provider_value = providers[providers_list.index(default_provider)]
                if default_provider_value[1] in configured_providers.keys():
                    if provider[:provider.rfind('_provider')] in configured_providers[default_provider_value[1]]:
                        providers.append((provider, default_provider_value[1]))

        for (option, value) in providers:
            try:
                domain.set_option(option, value)
            except NoOptionError:
                # If we come across an option that we don't recognize,
                # we should just ignore it and continue
                pass

        # Read in all the options from the configuration
        for opt in self.strip_comments_empty(self.options('domain/%s' % name)):
            if (opt['name'], opt['value']) not in providers:
                try:
                    domain.set_option(opt['name'], opt['value'])
                except NoOptionError:
                    # If we come across an option that we don't recognize,
                    # we should just ignore it and continue
                    pass

        # Determine if this domain is currently active
        domain.active = self.is_domain_active(name)

        return domain

    def new_domain(self, name):
        """
        Create a new, empty domain and return the SSSDDomain object for it.

        name:
          The name of the domain to create and return.

        === Returns ===
        The newly-created SSSDDomain object

        === Errors ===
        DomainAlreadyExistsError:
          The service being created already exists in the SSSDConfig object.
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError
        if self.has_section('domain/%s' % name):
            raise DomainAlreadyExistsError

        domain = SSSDDomain(name, self.schema)
        self.save_domain(domain)
        return domain

    def is_domain_active(self, name):
        """
        Is a particular domain set active

        name:
          The name of the configured domain to check

        === Returns ===
        True if the domain is active, False if it is inactive

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        NoDomainError:
          No domain by this name is configured
        """

        if not self.initialized:
            raise NotInitializedError

        if name not in self.list_domains():
            raise NoDomainError

        return name in self.list_active_domains()

    def activate_domain(self, name):
        """
        Activate a configured domain

        name:
          The name of the configured domain to activate

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        NoDomainError:
          No domain by this name is configured
        """

        if not self.initialized:
            raise NotInitializedError

        if name not in self.list_domains():
            raise NoDomainError

        item = self.get_option_index('sssd', 'domains')[1]
        if not item:
            self.set('sssd','domains', name)
            return

        # Turn the items into a set of dictionary keys
        # This guarantees uniqueness and makes it easy
        # to add a new value
        domain_dict = dict.fromkeys(striplist(item['value'].split(',')))
        if '' in domain_dict:
            del domain_dict['']

        # Add a new key for the domain being activated
        domain_dict[name] = None

        # Write out the joined keys
        self.set('sssd','domains', ", ".join(domain_dict.keys()))

    def deactivate_domain(self, name):
        """
        Deactivate a configured domain

        name:
          The name of the configured domain to deactivate

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        NoDomainError:
          No domain by this name is configured
        """

        if not self.initialized:
            raise NotInitializedError

        if name not in self.list_domains():
            raise NoDomainError
        item = self.get_option_index('sssd', 'domains')[1]
        if not item:
            self.set('sssd','domains', '')
            return

        # Turn the items into a set of dictionary keys
        # This guarantees uniqueness and makes it easy
        # to remove the one unwanted value.
        domain_dict = dict.fromkeys(striplist(item['value'].split(',')))
        if '' in domain_dict:
            del domain_dict['']

        # Remove the unwanted domain from the lest
        if name in domain_dict:
            del domain_dict[name]

        # Write out the joined keys
        self.set('sssd','domains', ", ".join(domain_dict.keys()))

    def delete_domain(self, name):
        """
        Remove a domain from the SSSDConfig object. This function will also
        remove this domain from the list of active domains in the [SSSD]
        section, if it is there.

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        """
        if not self.initialized:
            raise NotInitializedError

        # Remove the domain from the active domains list if applicable
        self.deactivate_domain(name)
        self.delete_option('section', 'domain/%s' % name)

    def save_domain(self, domain):
        """
        Save the changes made to the domain object back to the SSSDConfig
        object. If this domain is marked active, ensure it is present in the
        active domain list in the [SSSD] section

        domain_object:
          The SSSDDomain object to save to the configuration.

        === Returns ===
        No return value

        === Errors ===
        NotInitializedError:
          This SSSDConfig object has not had import_config() or new_config()
          run on it yet.
        TypeError:
          domain_object was not of type SSSDDomain
        """
        if not self.initialized:
            raise NotInitializedError
        if not isinstance(domain, SSSDDomain):
            raise TypeError

        name = domain.get_name()

        oldindex = None
        if domain.oldname and domain.oldname != name:
            # We are renaming this domain
            # Remove the old section

            self.deactivate_domain(domain.oldname)
            oldindex = self.delete_option('section', 'domain/%s' %
                                          domain.oldname)

            # Reset the oldname, in case we're not done with
            # this domain object.
            domain.oldname = None;

        sectionname = 'domain/%s' % name
        (no, section_subtree) = self.findOpts(self.opts, 'section', sectionname)

        if name not in self.list_domains():
            self.add_section(sectionname, []);

        section_options = self.options(sectionname)[:]
        for option in section_options:
            if option['type'] == 'option':
                if option['name'] not in domain.get_all_options():
                    self.delete_option_subtree(section_subtree['value'], 'option', option['name'], True)

        for option,value in domain.get_all_options().items():
            if (type(value) == list):
                value = ', '.join(value)
            if option == "debug_level":
                value = self._get_debug_level_val(value)
            self.set(sectionname, option, str(value))

        if domain.active:
            self.activate_domain(name)
        else:
            self.deactivate_domain(name)



""" Test cases for autofs responder

:requirement: AD Provider - automount
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""

from __future__ import print_function
import subprocess
import time
import pytest
from sssd.testlib.common.utils import sssdTools


@pytest.mark.usefixtures("enable_autofs_schema", "enable_autofs_service")
@pytest.mark.automount
class Testautofsresponder(object):
    """ Autofs responder test cases

    :setup:
      1. Configure nfs server exporting /export
      2. Load autofs maps to Active Directory under
         ou=automout,dc=<ad-domain>,dc=test
      3. Add an nisMapEntry in automount specifying nfs server,
         directory and file system(nfs)
      4. Join RHEL7 client to Windows AD using realm
    """
    @pytest.mark.parametrize('add_nisobject', ['/export'], indirect=True)
    @pytest.mark.tier1
    def test_001_searchbasedn(self, multihost, add_nisobject):
        """
        :title: IDM-SSSD-TC: AD-Provider Automount: Verify automount rules
         are searched from basedn
        :id: 48baf5ef-3757-45fd-954d-93c94264f880
        :steps:
          1. Edit sssd.conf and specify autofs_provider = ad and restart autofs
          2. Access /export share
        :expectedresults:
          1. Should succeed
          2. /export share should be mounted successfully
        """
        # pylint: disable=unused-argument
        multihost.master[0].run_command(['touch', '/export/nfs-test'])
        for service in ['sssd', 'autofs']:
            srv = 'systemctl restart %s' % service
            try:
                multihost.client[0].run_command(srv)
            except subprocess.CalledProcessError:
                pytest.fail("Unable to start %s service" % service)
            time.sleep(5)
        try:
            multihost.client[0].run_command(['automount', '-m'])
        except subprocess.CalledProcessError:
            pytest.fail("automount -m command failed")
        nfs_test = 'ls -l /export/nfs-test'
        cmd = multihost.client[0].run_command(nfs_test, raiseonerr=False)
        assert cmd.returncode == 0

    @pytest.mark.parametrize('add_nisobject', ['/export'], indirect=True)
    @pytest.mark.tier1
    def test_002_offline(self, multihost, add_nisobject):
        """
        :title: IDM-SSSD-TC: AD-Provider Automount: Verify automount
         maps are retrieved from cache when sssd is offline
        :id: 84f3c1ba-0321-4c3b-89c7-de90b4f32639
        :steps:
          1. Edit sssd.conf and specify autofs_provider = ad
          2. Restart Autofs
          3. pkill -USR1 sssd
          4. access /export/nfs-test
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. sssd should be offline
          4. /export/nfs-test share should be accessible from client
        """
        multihost.master[0].run_command(['touch', '/export/nfs-test'])
        for service in ['sssd', 'autofs']:
            srv = 'systemctl restart %s' % service
            try:
                multihost.client[0].run_command(srv)
            except subprocess.CalledProcessError:
                pytest.fail("Unable to start %s service" % service)
        time.sleep(30)
        automount = 'automount -m'
        try:
            multihost.client[0].run_command(automount)
        except subprocess.CalledProcessError:
            pytest.xfail("automount -m command failed")
        pkill_cmd = 'pkill -USR1 sssd'
        cmd = multihost.client[0].run_command(pkill_cmd, raiseonerr=False)
        assert cmd.returncode == 0
        try:
            multihost.client[0].run_command(automount)
        except subprocess.CalledProcessError:
            pytest.fail("automount -m command failed")
        nfstest = 'ls -l /export/nfs-test'
        cmd = multihost.client[0].run_command(nfstest, raiseonerr=False)
        assert cmd.returncode == 0

    @pytest.mark.parametrize('add_nisobject', ['/export'], indirect=True)
    @pytest.mark.tier1
    def test_003_setbasedn(self, multihost, set_autofs_search_base,
                           add_nisobject):
        """
        :title: IDM-SSSD-TC: AD-Provider Automount: Verify automount
         rules are searched when ldap_autofs_search_base is set
        :id: e8dbd94d-c557-4533-8ab7-bc891e1609a3
        :steps:
          1. Edit sssd.conf and specify below parameters: autofs_provider = ad
             ldap_autofs_search_base = ou=automount,dc=<ad-domain>
          2. Restart sssd
          3. Execute automount -m
          4. Access /export shared
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. automount -m should succeed and show maps from AD
          4. client is able mount nfs share /export
        """
        # pylint: disable=unused-argument
        journalctl_cmd = "journalctl -x -n 50 --no-pager"
        multihost.master[0].run_command(['touch', '/export/nfs-test'])
        for service in ['sssd', 'autofs']:
            srv = 'systemctl restart %s' % service
            try:
                multihost.client[0].run_command(srv)
            except subprocess.CalledProcessError:
                multihost.client[0].run_command(journalctl_cmd)
                pytest.fail("Unable to start %s service" % service)
        time.sleep(30)
        try:
            multihost.client[0].run_command(['automount', '-m'])
        except subprocess.CalledProcessError:
            pytest.fail("automount -m command failed")
        nfs_test = 'ls -l /export/nfs-test'
        cmd = multihost.client[0].run_command(nfs_test, raiseonerr=False)
        assert cmd.returncode == 0

    @pytest.mark.parametrize('add_nisobject', ['/export'], indirect=True)
    @pytest.mark.tier1
    def test_004_autofsnone(self, multihost, set_autofs_search_base,
                            add_nisobject):
        """
        :title: IDM-SSSD-TC: AD-Provider Automount: Verify maps are
         loaded from cache and maps are accessible when autofs_provider is None
        :id: 82e5a42d-a097-4604-bcd2-9e8af71bd511
        :steps:
          1. Set autofs_provider = None
          2. Restart sssd service
          3. Execute automount -m
          4. Run ls -l /export/nfs-share
        :expectedresults:
          1. Should succeed
          2. sssd service should be started successfully
          3. automount -m should execute successfully
          4. /export/nfs-share should be accessible
        """
        multihost.master[0].run_command(['touch', '/export/nfs-test'])
        for service in ['sssd', 'autofs']:
            srv = 'systemctl restart %s' % service
            try:
                multihost.client[0].run_command(srv)
            except subprocess.CalledProcessError:
                pytest.fail("Unable to start %s service" % service)
        time.sleep(30)
        try:
            multihost.client[0].run_command(['automount', '-m'])
        except subprocess.CalledProcessError:
            pytest.fail("automount -m command failed")
        nfs_test = 'ls -l /export/nfs-test'
        cmd = multihost.client[0].run_command(nfs_test, raiseonerr=False)
        assert cmd.returncode == 0
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        section = 'domain/{}'.format(domain_name)
        sssd_params = {'autofs_provider': 'none'}
        client.sssd_conf(section, sssd_params)
        srv = 'systemctl restart sssd'
        try:
            multihost.client[0].run_command(srv)
        except subprocess.CalledProcessError:
            pytest.fail("Unable to start sssd")
        time.sleep(10)
        try:
            multihost.client[0].run_command(['automount', '-m'])
        except subprocess.CalledProcessError:
            pytest.fail("automount -m command failed")
        cmd = multihost.client[0].run_command(nfs_test, raiseonerr=False)
        assert cmd.returncode == 0

    @pytest.mark.parametrize('add_nisobject', ['/export'], indirect=True)
    @pytest.mark.tier2
    def test_005_autofsnotset(self, multihost, set_autofs_search_base,
                              add_nisobject):
        """
        :title: IDM-SSSD-TC: AD-Provider Automount: Verify automount maps are
         loaded from Active Directory when autofs provider is not set
         in sssd.conf
        :id: a05ac41a-6668-4cde-8109-498baa5300b6
        :steps:
          1. Make sure autofs_provider is not set in sssd.conf,
          2. Access nfs share /export/nfs-test
        :expectedresults:
          1. autofs_provider is not set
          2. Verify automount maps are loaded from AD and client is able to
             mount nfs share
        """
        # pylint: disable=unused-argument
        client = sssdTools(multihost.client[0])
        domain_name = client.get_domain_section_name()
        section = 'domain/{}'.format(domain_name)
        sssd_params = {'autofs_provider': 'none'}
        client.sssd_conf(section, sssd_params, action='delete')
        journalctl_cmd = "journalctl -x -n 50 --no-pager"
        multihost.master[0].run_command(['touch', '/export/nfs-test'])
        for service in ['sssd', 'autofs']:
            restart = 'systemctl restart %s' % service
            cmd = multihost.client[0].run_command(restart, raiseonerr=False)
            if cmd.returncode != 0:
                multihost.client[0].run_command(journalctl_cmd)
                pytest.fail("Failed to restart %s" % service)
        automount = 'automount -m'
        cmd = multihost.client[0].run_command(automount, raiseonerr=False)
        assert cmd.returncode == 0
        nfstest = 'ls -l /export/nfs-test'
        cmd = multihost.client[0].run_command(nfstest, raiseonerr=False)
        assert cmd.returncode == 0

    @pytest.mark.parametrize('add_nisobject', ['/project1'], indirect=True)
    @pytest.mark.tier2
    def test_006_updatedmaps(self, multihost, set_autofs_search_base,
                             add_nisobject):
        """
        :title: IDM-SSSD-TC: AD-Provider Automount: Verify sssd properly
         updates cache when automount maps are updated with more entries
         in Active Directory
        :id: b0e28b05-468a-438c-bf8f-044f1ec73b05
        :steps:
          1. Add /project1 entry to auto.direct map
          2. Restart autofs
          3. Access /project1 nfs share
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Verify /project1 can be mounted on nfs client
        """
        # pylint: disable=unused-argument
        nfstest = '/project1/project1-test'
        create_nfsfile = 'touch %s' % nfstest
        multihost.master[0].run_command(create_nfsfile)
        for service in ['sssd', 'autofs']:
            restart = 'systemctl restart %s' % service
            cmd = multihost.client[0].run_command(restart, raiseonerr=False)
            if cmd.returncode != 0:
                journalctl_cmd = "journalctl -x -n 50 --no-pager"
                multihost.client[0].run_command(journalctl_cmd)
                pytest.fail("Failed to restart %s" % service)
        automount = 'automount -m'
        cmd = multihost.client[0].run_command(automount, raiseonerr=False)
        assert cmd.returncode == 0
        nfstest = 'ls -l %s' % nfstest
        cmd = multihost.client[0].run_command(nfstest, raiseonerr=False)
        assert cmd.returncode == 0



""" Automation of password policy test suite

:requirement: password_policy
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
"""
from __future__ import print_function
import re
import time
import pytest
from constants import ds_instance_name
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.ssh2_python import check_login_client_bool


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups',
                         'enable_password_check_syntax')
@pytest.mark.passwordcheck
class TestPasswordCheck(object):
    """
    This is test case class for password policy test suite
    """
    @pytest.mark.converted('test_ldap.py', 'test_ldap__change_password')
    @staticmethod
    @pytest.mark.tier2
    def test_0001_changeuserpass(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: password_policy: Change users
         password with ldap_pwmodify_mode in sssd.conf
        :id: dc70641a-bf1b-445c-926c-7ca693a87615
        :customerscenario: True
        """
        # Automation of BZ795044(rhel7) and BZ1695574(rhel8)
        # "exop" will change password with extended operation,
        # "ldap_modify" with direct modification of userPassword attribute.
        tools = sssdTools(multihost.client[0])
        user = 'foo9@%s' % ds_instance_name
        section = "domain/%s" % ds_instance_name
        for value in ['exop', 'ldap_modify']:
            domain_parameter = {'ldap_pwmodify_mode': value}
            tools.sssd_conf(section, domain_parameter, action='update')
            tools.clear_sssd_cache()
            time.sleep(5)
            change_pass = tools.change_user_password(user, 'Secret123',
                                                     'Secret123',
                                                     'bumblebee@123',
                                                     'bumblebee@123')
            assert change_pass == 3
            # Verify the login of user with updated password
            ssh = check_login_client_bool(multihost, user, 'bumblebee@123')

            # Revert back the password to old one
            change_pass_old = tools.change_user_password(user, 'bumblebee@123',
                                                         'bumblebee@123',
                                                         'Secret123',
                                                         'Secret123')
            assert ssh, f'{user} is not able to login.'
            assert change_pass_old == 3

    @pytest.mark.converted('test_ldap.py', 'test_ldap__change_password_new_pass_not_match')
    @staticmethod
    @pytest.mark.tier2
    def test_0002_newpassnotmatch(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: password_policy: New password is
         not matching with retype password with ldap_pwmodify_mode in sssd.conf
        :id: ef5a83f3-4560-46dd-b7e7-c6bbdf0da551
       """
        # Automation of BZ795044(rhel7) and BZ1695574(rhel8)
        # "exop" will change password with extended operation,
        # "ldap_modify" with direct modification of userPassword attribute.
        tools = sssdTools(multihost.client[0])
        user = 'foo9@%s' % ds_instance_name
        section = "domain/%s" % ds_instance_name
        rm_secure_log = 'echo > /var/log/secure'
        for value in ['exop', 'ldap_modify']:
            multihost.client[0].run_command(rm_secure_log)
            domain_parameter = {'ldap_pwmodify_mode': value}
            tools.sssd_conf(section, domain_parameter, action='update')
            tools.clear_sssd_cache()
            change_pass = tools.change_user_password(user, 'Secret123',
                                                     'Secret123',
                                                     'bumblebee@123',
                                                     'bumblebee')
            assert change_pass == 5

    @pytest.mark.converted('test_ldap.py', 'test_ldap__change_password_lowercase')
    @staticmethod
    @pytest.mark.tier2
    def test_0003_smallnewpass(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: password_policy: Check new
         password quality check with ldap_pwmodify_mode in sssd.conf
        :id: 00c39f98-3420-4e95-ac96-0929fe771eff
       """
        # Automation of BZ795044(rhel7) and BZ1695574(rhel8)
        # "exop" will change password with extended operation,
        # "ldap_modify" with direct modification of userPassword attribute.
        # BZ1795220(rhel8.3)
        tools = sssdTools(multihost.client[0])
        user = 'foo9@%s' % ds_instance_name
        section = "domain/%s" % ds_instance_name
        rm_secure_log = 'echo > /var/log/secure'
        for value in ['exop', 'ldap_modify']:
            multihost.client[0].run_command(rm_secure_log)
            domain_parameter = {'ldap_pwmodify_mode': value}
            tools.sssd_conf(section, domain_parameter, action='update')
            tools.clear_sssd_cache()
            change_pass = tools.change_user_password(user, 'Secret123',
                                                     'Secret123', 'red_32',
                                                     'red_32')
            assert change_pass == 4
            log1 = re.compile(r'pam_sss.passwd:chauthtok.:\sUser\sinfo\s'
                              r'message:\sPassword\schange\sfailed.*')
            time.sleep(5)
            test_str_log = multihost.client[0].get_file_contents(
                '/var/log/secure')
            assert log1.search(test_str_log.decode())

    @pytest.mark.converted('test_ldap.py', 'test_ldap__change_password_wrong_current')
    @staticmethod
    @pytest.mark.tier2
    def test_0004_wrongcurrentpass(multihost):
        """
        :title: IDM-SSSD-TC: ldap_provider: password_policy: Check wrong
         current password with ldap_pwmodify_mode in sssd.conf
        :id: 12ce875a-d8b9-4165-980a-f649459f45f0
        """
        # Automation of BZ795044(rhel7) and BZ1695574(rhel8)
        # "exop" will change password with extended operation,
        # "ldap_modify" with direct modification of userPassword attribute.
        tools = sssdTools(multihost.client[0])
        user = 'foo9@%s' % ds_instance_name
        section = "domain/%s" % ds_instance_name
        rm_secure_log = 'echo > /var/log/secure'
        for value in ['exop', 'ldap_modify']:
            multihost.client[0].run_command(rm_secure_log)
            domain_parameter = {'ldap_pwmodify_mode': value}
            tools.sssd_conf(section, domain_parameter, action='update')
            tools.clear_sssd_cache()
            change_pass = tools.change_user_password(user, 'Secret123',
                                                     'secret@123',
                                                     'redhat@321',
                                                     'redhat@321')
            assert change_pass == 6
            log1 = re.compile(r'Password\schange\sfailed.\sServer\smessage:\s'
                              r'Old\spassword\snot\saccepted.')
            time.sleep(5)
            test_str_log = multihost.client[0].get_file_contents(
                '/var/log/secure')
            assert log1.search(test_str_log.decode())



"""Automation poor man's backtrace

:requirement: Poor Man's Backtrace
:casecomponent: sssd
:subsystemteam: sst_idm_sssd
:upstream: yes
:status: approved
:bugzilla: https://bugzilla.redhat.com/show_bug.cgi?id=1949149
"""

from __future__ import print_function
import re
import time
import pytest
from sssd.testlib.common.utils import sssdTools
from sssd.testlib.common.ssh2_python import check_login_client
from constants import ds_instance_name


def bad_url(multihost):
    tools = sssdTools(multihost.client[0])
    hostname = multihost.master[0].sys_hostname
    section = f"domain/{ds_instance_name}"
    bad_ldap_uri = f"ldaps://typo.{hostname}"
    domain_params = {'ldap_uri': bad_ldap_uri}
    tools.sssd_conf(section, domain_params)


def no_fallback_dir(multihost):
    tools = sssdTools(multihost.client[0])
    section = f"domain/{ds_instance_name}"
    domain_params = {'fallback_homedir': ''}
    tools.sssd_conf(section, domain_params, action='delete')
    tools.clear_sssd_cache()
    user = f'foo1@{ds_instance_name}'
    # Authenticate user
    check_login_client(multihost, user, 'Secret123')


@pytest.mark.usefixtures('setup_sssd', 'create_posix_usersgroups')
@pytest.mark.backtrace
class TestPoorManBacktrace(object):
    """ Check sssd backtrace feature """
    @staticmethod
    @pytest.mark.tier2
    def test_0001_bz2021196(multihost, backupsssdconf):
        """
        :title: avoid duplicate backtraces
        :id: d4d8a0a0-ab90-4c8f-8087-95dc7ad3f3ae
        :customerscenario: true
        :bugzilla:
          https://bugzilla.redhat.com/show_bug.cgi?id=2021196
          https://bugzilla.redhat.com/show_bug.cgi?id=2021499
        :steps:
          1. Modify sssd.conf with a typo in the ldap_uri
          2. Remove debug_level from domain section
          3. Start sssd
          4. Truncate domain logs
          5. Lookup any user
          6. Check logs generate a backtrace
          7. Lookup same user again
          8. Check logs dont have repeated backtrace
        :expectedresults:
          1. Bad url successfully added in sssd.conf
          2. debug_level not set in sssd.conf
          3. Should succeed
          4. Should succeed
          5. Should fail as expected
          6. Should succeed
          7. Should fail as expected
          8. Should have string 'skipping repetitive backtrace'
        """
        arch = multihost.client[0].run_command(
            'uname -m', raiseonerr=False).stdout_text
        if 'x86_64' not in arch:
            pytest.skip("Test is unstable on architectures other than x86_64.")
        bad_url(multihost)
        tools = sssdTools(multihost.client[0])
        section = f"domain/{ds_instance_name}"
        domain_params = {'debug_level': ''}
        tools.sssd_conf(section, domain_params, action='delete')
        tools.clear_sssd_cache()
        logfile = f'/var/log/sssd/sssd_{ds_instance_name}.log'
        log_str = multihost.client[0].get_file_contents(logfile). \
            decode('utf-8')
        print(f'\n{logfile}\n+===++++++===+\n{log_str}\n')
        time.sleep(2)
        # Truncate logs to generate fresh backtrace on user lookup and compare
        # with the logs generated with the second lookup for same error
        multihost.client[0].run_command('sssctl logs-remove')
        cmd = f'getent passwd fakeuser@{ds_instance_name}'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        time.sleep(2)
        msg = 'BACKTRACE DUMP ENDS HERE'
        msg2 = '... skipping repetitive backtrace ...'
        pattern = re.compile(fr'{msg}')
        pattern2 = re.compile(fr'{msg2}')
        log_str1 = multihost.client[0].get_file_contents(logfile). \
            decode('utf-8')
        print(f'\n{logfile}\n+===++++++===+\n{log_str1}\n')
        multihost.client[0].run_command('sssctl logs-remove')
        multihost.client[0].run_command(cmd, raiseonerr=False)
        time.sleep(2)
        log_str2 = multihost.client[0].get_file_contents(logfile). \
            decode('utf-8')
        print(f'\n{logfile}\n+===++++++===+\n{log_str2}\n')
        # Check the backtrace is dumped first time and no backtrace is skipped
        assert len(pattern.findall(log_str1)) == len(pattern2.findall(log_str1)) == 1
        # Check there is no new backtrace with the same issue and repeative
        # backtrace is skipped
        assert pattern2.search(log_str2) and not pattern.search(log_str2)

    @staticmethod
    @pytest.mark.tier1_2
    def test_0002_bz1949149(multihost, backupsssdconf):
        """
        :title: backtrace is disabled if debug level >= 9
        :id: 50f2d501-3296-4229-86a0-b81844381637
        :steps:
          1. Set debug_level to 9 in all sections of sssd.conf
          2. Dont set fallback_homedir
          3. Start sssd
          4. Login as user
          5. Set bad url for ldap_uri
          6. Restart sssd
          7. Check logs
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Login succeeds
          5. Should succeed
          6. Should succeed
          7. Logs should not have string 'BACKTRACE'
        """
        section = f"domain/{ds_instance_name}"
        param = {'debug_level': '9'}
        serv_list = ['sssd', section, 'nss', 'pam']
        tools = sssdTools(multihost.client[0])
        for serv in serv_list:
            tools.sssd_conf(serv, param)
        no_fallback_dir(multihost)
        bad_url(multihost)
        tools.service_ctrl('restart', 'sssd')
        cmd = f'getent passwd fakeuser@{ds_instance_name}'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        log_list = ['sssd', f'sssd_{ds_instance_name}',
                    'sssd_nss', 'sssd_pam']
        for logfilename in log_list:
            log = f'/var/log/sssd/{logfilename}.log'
            log_str = multihost.client[0].get_file_contents(log).decode(
                'utf-8')
            find = re.compile(r'BACKTRACE DUMP ENDS HERE')
            assert not find.search(log_str)

    @staticmethod
    @pytest.mark.tier1_2
    def test_0003_bz1949149(multihost, backupsssdconf):
        """
        :title: set debug_backtrace_enabled false
        :id: b8084e03-5e21-45ee-a463-65ab537fa110
        :steps:
          1. Set debug_backtrace_enabled to false in sssd.conf
          2. Start sssd
          3. Login as user
          4. Set bad url for ldap_uri
          5. Restart sssd
          6. Lookup any user
          7. Check logs
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Login succeeds
          4. Should succeed
          5. Should succeed
          6. Should fail as expected
          7. Logs dont have backtrace
        """
        section = f"domain/{ds_instance_name}"
        param = {'debug_backtrace_enabled': 'false'}
        serv_list = ['sssd', section, 'nss', 'pam']
        tools = sssdTools(multihost.client[0])
        for serv in serv_list:
            tools.sssd_conf(serv, param)
        no_fallback_dir(multihost)
        bad_url(multihost)
        tools.service_ctrl('restart', 'sssd')
        cmd = f'getent passwd fakeuser@{ds_instance_name}'
        multihost.client[0].run_command(cmd, raiseonerr=False)
        log_list = ['sssd', 'sssd_example1', 'sssd_nss', 'sssd_pam']
        for logfilename in log_list:
            log = f'/var/log/sssd/{logfilename}.log'
            log_str = multihost.client[0].get_file_contents(log).decode(
                'utf-8')
            find = re.compile(r'BACKTRACE DUMP ENDS HERE')
            assert not find.search(log_str)

    @staticmethod
    @pytest.mark.tier1_2
    def test_0004_bz1949149(multihost, backupsssdconf):
        """
        :title: backtrace level is 0 with debug level set to 0
        :id: 4376d596-a613-447c-8f85-e3f3fbc05728
        :steps:
          1. Set debug_level to 0
          2. Remove fallback dir from sssd.conf
          3. Login as a user
          4. Add bad url to ldap_uri in sssd.conf
          5. Restart sssd
          6. Lookup any user
          7. Check logs has backtraces for step 3 and 6
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          7. Backtrace generated only for level 0 logs
        """
        section = f"domain/{ds_instance_name}"
        param = {'debug_level': '0'}
        serv_list = ['sssd', section, 'nss', 'pam']
        tools = sssdTools(multihost.client[0])
        for serv in serv_list:
            tools.sssd_conf(serv, param)
        no_fallback_dir(multihost)
        bad_url(multihost)
        tools.service_ctrl('restart', 'sssd')
        log_list = ['sssd', f'sssd_{ds_instance_name}', 'sssd_nss', 'sssd_pam']
        find1 = re.compile(r'BACKTRACE DUMP ENDS HERE')
        find2 = re.compile(r'.0x0010.')
        for logfilename in log_list:
            log = f'/var/log/sssd/{logfilename}.log'
            log_str = multihost.client[0].get_file_contents(log). \
                decode('utf-8')
            log_lines = log_str.splitlines()
            # Check only error of level 0x0010 generates backtrace
            for index, line in enumerate(log_lines):
                if find1.search(line):
                    log_level = log_lines[index - 1]
                    assert find2.search(log_level)

    @staticmethod
    @pytest.mark.tier1_2
    def test_0005_bz1949149(multihost, backupsssdconf):
        """
        :title: backtrace level is 1 with debug level set to 1
        :id: 8a8adcdd-63bc-4a64-83cd-5c7b76fe745a
        :steps:
          1. Set debug_level to 1
          2. Remove fallback dir from sssd.conf
          3. Login as a user
          4. Add bad url to ldap_uri in sssd.conf
          5. Restart sssd
          6. Lookup any user
          7. Check logs has backtraces for step 3 and 6
        :expectedresults:
          1. Should succeed
          2. Should succeed
          3. Should succeed
          4. Should succeed
          5. Should succeed
          6. Should succeed
          7. Backtrace is generated only for level 0 and level 1 logs
        """
        section = "domain/{ds_instance_name}"
        param = {'debug_level': '1'}
        serv_list = ['sssd', section, 'nss', 'pam']
        tools = sssdTools(multihost.client[0])
        for serv in serv_list:
            tools.sssd_conf(serv, param)
        no_fallback_dir(multihost)
        bad_url(multihost)
        tools.service_ctrl('restart', 'sssd')
        log_list = ['sssd', f'sssd_{ds_instance_name}', 'sssd_nss', 'sssd_pam']
        find1 = re.compile(r'BACKTRACE DUMP ENDS HERE')
        find2 = re.compile(r'.0x0010.')
        find3 = re.compile(r'.0x0020.')
        for logfilename in log_list:
            log = f'/var/log/sssd/{logfilename}.log'
            log_str = multihost.client[0].get_file_contents(log). \
                decode('utf-8')
            log_lines = log_str.splitlines()
            # Check only error of level 0x0010 and 0x0020 generates backtrace
            for index, line in enumerate(log_lines):
                if find1.search(line):
                    log_level = log_lines[index - 1]
                    assert find2.search(log_level) or find3.search(log_level)
